/* * * * * * * * * * * *  PIETER SIMOONS' LIBRARY  * * * * * * * * * * * *\
 *                                                                       *
 * Note that unless you include one of the graphical libraries (such as  *
 * PVGA.H), the Drawing, Sprites and morph functions won't work.         *
 *                                                                       *
 * * * * * * * * * * * * * * GENERIC FUNCTIONS * * * * * * * * * * * * * *
 * sgn (num) [int macro]                                                 *
 *   returns the sign of the number (-1, 0 or 1)                         *
 * getvect (intr) [intpointer extern] | setvect (intr,adr) [void extern] *
 *   get or set an interrupt vector [Dos.H]                              *
 * inp (port) [char alias extern] | outp (port, num) [void alias extern] *
 *   read and write data from a hardware port [Conio.H]                  *
 * mspiral | mrandm () [void macro]                                      *
 *   creates a char[128] that can be used as morphtype                   *
 *                                                                       *
 * * * * * * * * * * * * * *  INITIALIZATIONS  * * * * * * * * * * * * * *
 * txtmode () [void macro]                                               *
 *   returns to text mode 03h                                            *
 *                                                                       *
 * findgfx () [char/boolean extern]                                      *
 *   detects whether the mode is supported or not                        *
 * gfxmode () [void extern]                                              *
 *   initializes graphics mode and clears the screen                     *
 *                                                                       *
 * * * * * * * * * * * * * * * PAGE FLIPPING * * * * * * * * * * * * * * *
 * vwait () [void asm]                                                   *
 *   waits for the vertical retrace interrupt                            *
 * morph (morphtype) [void]                                              *
 *   slowly changes the visual page into the active page using morphtype *
 *                                                                       *
 * flip () [void extern]                                                 *
 *   moves the contents of the active page onto the visual page          *
 *                                                                       *
 * * * * * * * * * * * * * COLORS AND PALETTES * * * * * * * * * * * * * *
 * pal (c,r,g,b) [void macro]                                            *
 *   sets the red, green and blue values of color c to r, g, and b       *
 * pget (c,r,g,b) [void macro]                                           *
 *   returns the red, green and blue values of color c in r, g and b     *
 * setpal (r,g,b | colorarray) [void]                                    *
 *   instantly sets all colors on screen to the specified values         *
 * fadeover (r,g,b | colorarray) [void]                                  *
 *   fades all colors on screen to the specified red, green, blue values *
 * cycle (number) [void]                                                 *
 *   cycles a number of groups of eight colors, simulating animation     *
 *                                                                       *
 * * * * * * * * * * * * * * * *  DRAWING  * * * * * * * * * * * * * * * *
 * line (x1,y1,x2,y2,color) [void]                                       *
 *   uses the Bresenham algorithm to draw a line between two points      *
 * box (x1,y1,x2,y2,color) [void]                                        *
 *   creates a cube with corners (x1,y1) and (x2,y2) filled with color c *
 * xbox (x1,y1,x2,y2,color) [void]                                       *
 *   creates a similar cube, only XORs c to the colors already there     *
 * rect (x1,y1,x2,y2,color) [void]                                       *
 *   creates a rectangle, i.e. doesn't fill the cube                     *
 * crect (x1,y1,x2,y2,color) [void]                                      *
 *   clips the rectangle, i.e. doesn't draw any off-screen parts         *
 *                                                                       *
 * pset (x,y,color) [void extern]                                        *
 *   gives the pixel at coordinates (x,y) the specified color            *
 * dpset (x,y,color) [void extern]                                       *
 *   puts a pixel directly on the screen instead of in the flip buffer   *
 * xpset (x,y,color) [void extern]                                       *
 *   XORs a pixel at the given coordinates with the given color          *
 * point (x,y) [unsigned char extern]                                    *
 *   returns the color of the pixel at coordinates (x,y)                 *
 * cls (color) [void extern]                                             *
 *   fills the entire screen with the color                              *
 *                                                                       *
 * * * * * * * * * * * * * * * *  SPRITES  * * * * * * * * * * * * * * * *
 * checksize (x1,y1,x2,y2,array) [char/boolean macro]                    *
 *   checks whether the sprite in the rectangle would fit into the array *
 * sget (x1,y1,x2,y2,array) [void]                                       *
 *   saves screen data from the specified rectangle in the array	 *
 * sput | sputt | sputb | sputx (x,y,array) [void]                       *
 *   puts data on screen using Set, Transparent, Background or XOR       *
 * sputs (x,y,array,color) [void]                                        *
 *   puts a shadow of the sprite array on screen, in the given color     *
 * sput18 (x,y,array) [void]                                             *
 *   puts data on screen, assuming size to be 18x18 pixels               *
 *                                                                       *
 * * * * * * * * * * * * * * * * KEYBOARDS * * * * * * * * * * * * * * * *
 * [global var] keydown [unsigned char [128]]                            *
 *   each element holds (128) if the key is down, and (0) if it's up     *
 * kbsetup | kbrestore () [void macro]                                   *
 *   initialize and de-initialize the new keyboard handler               *
 * kbnew | kbold (...) [void interrupt]                                  *
 *   old and new keyboard interrupt handlers                             *
 * leds (leds) [void macro]                                              *
 *   toggles caps lock (64), num lock (32) and/or scroll lock (16)       *
 * minit | mshow | mhide [void macro]                                    *
 *   initializes, shows or hides the mouse cursor                        *
 * mstat(x,y,b) [void macro]                                             *
 *   gets mouse cursor x and y position, and button status               *
 * mbind(l,u,r,d) [void macro]                                           *
 *   binds mouse cursor to the area given by left, right, up and down    *
 *                                                                       *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

// ========================= GENERIC FUNCTIONS =========================
typedef unsigned char uchar;
#define sgn(x) ((x)?((signed)(x)>0?1:-1):0)
#define min(x,y) (x>y?y:x)
#define max(x,y) (x>y?x:y)
#include <dos.h>
/*extern "C" {
//  void  interrupt (far *cdecl getvect (int)) (...);
//  void  setvect   (int, void interrupt (far *) (...));
  uchar inportb   (unsigned);
  void  outportb  (unsigned, unsigned char);
  int   rand      (void);
  void  srand     (unsigned);
  long  time      (long*);
}
inline int random (int n) { return (int) (((long) rand() * n) / 0x8000u); }
#ifndef inp
#define inp(portid)    inportb  (portid)
#define outp(portid,v) outportb (portid,v)
#endif  */

#define mspiral char spiral[128] = \
  {0,1,2,3,4,5,6,7,7,7,7,7,7,7,7,6,5,4,3,2,1,0,0,0,0,0,0,0,1,2,3,4,\
   5,6,6,6,6,6,6,5,4,3,2,1,1,1,1,1,2,3,4,5,5,5,5,4,3,2,2,2,3,4,4,3,\
   0,0,0,0,0,0,0,0,1,2,3,4,5,6,7,7,7,7,7,7,7,7,6,5,4,3,2,1,1,1,1,1,\
   1,1,2,3,4,5,6,6,6,6,6,6,5,4,3,2,2,2,2,2,3,4,5,5,5,5,4,3,3,3,4,4}
#define mrandm char randm[128] = \
  {1,2,6,0,1,7,5,1,3,5,6,2,6,5,6,4,5,4,2,5,1,5,0,2,4,0,3,3,3,3,2,2,\
   3,3,4,6,3,1,6,5,1,4,7,2,7,4,7,6,2,6,1,0,5,7,4,0,7,7,4,0,1,0,7,0,\
   6,4,1,3,1,2,3,4,6,1,4,2,3,0,2,2,4,4,6,6,7,5,2,5,1,4,5,2,4,3,1,0,\
   7,0,7,6,1,3,5,2,5,5,1,7,3,3,4,7,3,0,0,5,7,6,0,6,5,0,6,0,2,1,7,7}
#define mdiam char diam[128] = \
  {3,4,3,2,3,5,4,3,2,1,2,3,4,6,5,4,3,2,1,0,1,2,3,4,5,7,6,5,4,3,2,1,\
   0,0,1,2,4,5,6,7,6,5,4,2,1,0,0,1,5,6,7,7,6,5,1,0,0,6,7,7,6,0,7,7,\
   3,3,4,3,2,3,4,5,4,3,2,1,2,3,4,5,6,5,4,3,2,1,0,1,2,3,4,5,6,7,6,5,\
   4,2,1,0,0,1,2,4,5,6,7,7,6,5,1,0,0,1,2,5,6,7,7,6,0,0,1,6,7,7,0,7}


// ========================== INITIALIZATIONS ==========================
// char/boolean findgfx () extern
// void gfxmode () extern
// void txtmode () extern

// =========================== PAGE FLIPPING ===========================
void vwait (void) {
      asm   mov dx, 0x03DA
  v1: asm { in  al, dx
	    and al, 8
	    jz  v1 }
  v2: asm { in  al, dx
	    and al, 8
	    jz  v2 }
}
#ifdef pset
void morph (char* mx) {
  unsigned int i, j, tx, ty;
  char c, n;
  for (n=0; n<64; n++) {
    vwait;
    for (i=0; i<MAXX; i+=8)
      for (j=0; j<MAXY; j+=8) {
	tx = i + mx[n]; ty = j + mx[n+64]; c = point (tx, ty);
        dpset (tx, ty, c);
      }
  }
}
#endif
// void flip () extern


// ======================= COLORS AND PALETTES =========================
#define pal(c,r,g,b) { outp(0x3c8,c);\
  outp(0x3c9, r); outp(0x3c9, g); outp(0x3c9, b); }
#define pget(c,r,g,b) { outp(0x3c7,c);\
  r = inp(0x3c9); g = inp(0x3c9); b = inp(0x3c9); }
void setpal (char tr, char tg, char tb) {
  int i; vwait;
  for (i = 0; i < 256; i++) { pal (i, tr, tg, tb); }
}
void setpal (uchar* array) {
  int i; vwait;
  for (i = 0; i < 256; i++) {
    pal (i, array [i*3], array [i*3+1], array [i*3+2]);
  }
}
void setpal (char* array) { setpal ((uchar*) array); }
void pclear (void) {
  int i; vwait;
  for (i = 0; i < 256; i++) pal (i, 0, 0, 0);
}
void fadeover (char tr, char tg, char tb) {
  int j;
  char i, r, g, b;
  for (i = 0; i < 64; i++) {
    vwait;
    for (j = 0; j < 256; j++) {
      pget (j, r, g, b);
      r += sgn (tr - r); g += sgn (tg - g); b += sgn (tb - b);
      pal (j, r, g, b);
    }
  }
}
void fadeover (char* array) {
  int j;
  char i, r, g, b;
  for (i = 0; i < 64; i++) {
    vwait;
    for (j = 0; j < 256; j++) {
      pget (j, r, g, b);
      r += sgn (array[j*3] - r); g += sgn (array[j*3+1] - g);
      b += sgn (array[j*3+2] - b);
      pal (j, r, g, b);
    }
  }
}
void fadeout (void) {
  int j;
  char i, r, g, b;
  for (i = 0; i < 64; i++) {
    vwait;
    for (j = 0; j < 256; j++) {
      pget (j, r, g, b);
      if (r) r--; if (g) g--; if (b) b--;
      pal (j, r, g, b);
    }
  }
}
void cycle (int n) {
  char i, j, r, g, b, r0, g0, b0;
  n = 256 - (n * 8);
  for (i=n; i; i+=8) {
    vwait;
    pget (i+7, r0, g0, b0);
    for (j=i+6; j>=i; j--) {
      pget (j, r, g, b); pal (j+1, r, g, b);
    }
    pal (i, r0, g0, b0);
  }
}


// ============================== DRAWING ==============================
#ifdef pset
void line (int x1, int y1, int x2, int y2, char color) {
  int xUnit, yUnit, yDiff = y2-y1, xDiff = x2-x1, error_term = 0, length, i;
  if (yDiff < 0) {
    yDiff = -yDiff;
    yUnit = -1;
  } else
    yUnit = 1;
  if (xDiff < 0) {
    xDiff = -xDiff;
    xUnit = -1;
  } else
    xUnit = 1;
  if (xDiff > yDiff) {
    length = xDiff+1;
    for (i=0; i<length; i++) {
      pset(x1,y1,color);
      x1 += xUnit;
      error_term += yDiff;
      if(error_term > xDiff) {
	error_term -= xDiff;
	y1 += yUnit;
      }
    }
  } else {
    length = yDiff+1;
    for (i=0; i<length; i++) {
      pset(x1,y1,color);
      y1 += yUnit;
      error_term += xDiff;
      if (error_term>0) {
	error_term -= yDiff;
	x1 += xUnit;
      }
    }
  }
}
void box (int x1, int y1, int x2, int y2, char color) {
  int i, j;
  for (i=y1; i<=y2; (y1<y2 ? i++ : i--))
    for (j=x1; j<=x2; (x1<x2 ? j++ : j--))
      pset (j, i, color);
}
void xbox (int x1, int y1, int x2, int y2, char color) {
  int i, j;
  for (i=y1; i<=y2; (y1<y2 ? i++ : i--))
    for (j=x1; j<=x2; (x1<x2 ? j++ : j--))
      xpset (j, i, color);
}
void rect (int x1, int y1, int x2, int y2, char color) {
  int i;
  for (i=y1; i<=y2; (y1<y2 ? i++ : i--)) {
    pset (x1, i, color); pset (x2, i, color); }
  for (i=x1; i<=x2; (x1<x2 ? i++ : i--)) {
    pset (i, y1, color); pset (i, y2, color); }
}
void crect (int x1, int y1, int x2, int y2, char color) {
  int i;
  if (x1 < 0) x1 = 0; if (y1 < 0) y1 = 0;
  if (x2 > 319) x2 = 319; if (y2 > 199) y2 = 199;
  for (i=y1; i<=y2; (y1<y2 ? i++ : i--)) {
    xpset (x1, i, color); xpset (x2, i, color); }
  for (i=x1; i<=x2; (x1<x2 ? i++ : i--)) {
    xpset (i, y1, color); xpset (i, y2, color); }
}
#endif
// void pset (int x, int y, char color) extern
// void dpset (int x, int y, char color) extern
// void xpset (int x, int y, char color) extern
// unsigned char point (int x, int y) extern
// void cls (char color) extern


// ============================= SPRITES ===============================
#ifdef pset
#define checksize(xa,ya,xb,yb,spr) \
  ((((xb-xa+1)*(yb-ya+1)+2)<=sizeof(spr))|((xb-xa)<255))
void sget (int x1, int y1, int x2, int y2, char* sprite) {
  char i, j, xsize, ysize, xsizeb;
  if (x1 > x2) { i = x1; x1 = x2; x2 = i; }
  if (y1 > y2) { i = y1; y1 = y2; y2 = i; }
  sprite[0] = xsize = x2-x1; sprite[1] = ysize = y2-y1; xsizeb = xsize + 1;
  for (i=0; i<=ysize; i++)
    for (j=0; j<=xsize; j++)
      sprite [i*xsizeb+j+2] = point (j+x1, i+y1);
}
void sput18 (int x1, int y1, char *sprite) {
  int x2 = x1;
  char i, j, xsize = 17, ysize = 17, c;
  for (i=0; i<=ysize; i++) {
    for (j=0; j<=xsize; j++) {
      c = sprite [i*xsize+i+j];
      pset (x2, y1, c);
      x2++;
    }
    y1++; x2 = x1;
  }
}
#define gensprlbeg                                    \
  int x2 = x1;                                        \
  char i, j, xsize = sprite[0], ysize = sprite[1], c; \
  for (i=0; i<=ysize; i++) {                          \
    for (j=0; j<=xsize; j++) {
      #define gensprlend                              \
      x2++;                                           \
    }                                                 \
    y1++; x2 = x1;                                    \
  }
void sput (int x1, int y1, char *sprite) {
  gensprlbeg
      c = sprite [i*xsize+i+j+2];
      pset (x2, y1, c);
  gensprlend
}
void sputt (int x1, int y1, char *sprite) {
  gensprlbeg
      c = sprite [i*xsize+i+j+2];
      if (c) pset (x2, y1, c);
  gensprlend
}
void sputb (int x1, int y1, char *sprite) {
  gensprlbeg
      if (!point (x2, y1)) {
	c = sprite [i*xsize+i+j+2];
	pset (x2, y1, c);
      }
  gensprlend
}
void sputx (int x1, int y1, char *sprite) {
  gensprlbeg
      c = sprite [i*xsize+i+j+2] ^ point (x2, y1);
      pset (x2, y1, c);
  gensprlend
}
void sputs (int x1, int y1, char *sprite, char col) {
  gensprlbeg
       if (sprite [i*xsize+i+j+2]) pset (x2, y1, col);
  gensprlend
  c=0; c++; // this to prevent compiler warning for unused variable
}
#endif

// ============================ KEYBOARDS ==============================
// ESC     1      F1   59-68   F11 87-88   SCRLLK 70   NUMLK  69
// `~     41      1     2-14   BACKSP 14               *      55
// TAB    15-28                ENTER  28               789-   71-74
// CAPSLK 58      A    30-40   "'     40               456+   75-78
// LSHIFT 42-54                RSHIFT 54               1230.  79-83
// LCTRL  29      LALT 56      SPACE  57
//

unsigned char keydown[128];
static void interrupt (far *kbold) (...);// = getvect (9);
void interrupt kbnew (...) {
  asm {
    cli;
  } kbn1: asm {
    in  al, 0x60;
    cmp al, 0xe0; jne kbn2;
    mov al, 0x20; out 0x20, al;
    jmp short kbn1;
  } kbn2: asm {
    mov bl, al; and bx, 127;
    and al, 128; xor al, 128; shr al, 7;
    mov	bx [keydown], al;
    mov al, 0x20; out 0x20, al;
    sti;
  }
}
void inline kbsetup   (void) { kbold = getvect (9); setvect (9, kbnew); }
void inline kbrestore (void) { setvect (9, kbold); }
#define leds(num) *((char far*)((void near*) 0x417)) = num
#define CAPS 64
#define NUM 32
#define SCROLL 16

#define minit asm { mov ax,0; int 51; }
#define mshow asm { mov ax,1; int 51; }
#define mhide asm { mov ax,2; int 51; }
#define mstat(x,y,b) asm { mov ax,3; int 51; \
  shr cx,1; mov x,cx; mov y,dx; mov b,bx;    }
#define mbind(l,u,r,d) asm { mov ax,7; mov cx,l; shl cx,1; mov dx,r; \
  shl dx,1; int 51; mov ax,8; mov cx,u; mov dx,d; int 51;            }
